-- dining philosophers paradigm.

-- The chopsticks are allocated as many as available 
-- and the allocation is completed as soon as the missing chopsticks are released.
 
-- JAVA implementation simulated with an Ada protected object 
-- simulates Java concurrency semantics
-- the Java wait operation is here an entry call with a barrier
-- the Java synchronized method simulated by Get_LR 
-- allocates two sticks if they are available
-- else allocate left stick if possible 
-- waits if both sticks have not been allocated

-- If the right stick is not available for philosopher C it is BookedRight(C + 1)
-- a left stick for philosopher C formerly BookedRight(C) by C - 1 is not allocated,
--  this avoids deadlock

-- If a left stick is not available for philosopher C it is BookedLeft(C)
-- a right stick for philosopher C formerly BookedLeft(C + 1) is not allocated
-- this avoids starvation
-- however this reintroduce deadlock unless philosopher C is denied access 
-- when one of its neighbour is waiting i.e. BookedRight(C) or BookedLeft(C + 1)

-- instrumented for concurrency denials ratio measurement

-- ********    reliable, deadlock free

with Global; use global;
package body Chop is

	type SticState is array(Id) of Boolean;
	subtype Cardinal is Integer range 0..2;
	
	protected Sticks is
		entry Get_LR(C : Id; Score : in out Cardinal); 
		procedure Release(C : in Id);
		-- Score is the number of already allocated sticks
	private 
		entry Wait(C : Id; Score : in out Cardinal);
		Available: SticState := (others => True);
		BookedRight : SticState := (others => False);
		BookedLeft : SticState := (others => False);
		NotifyAll : Boolean := False; -- Java signal simulation  
	end Sticks;
	
    protected body Sticks is
   
      entry  Get_LR(C : Id; Score : in out Cardinal) when True is
	  begin
		 case Score is
         when 0 => if not BookedRight(C) and not BookedLeft(C + 1) then
					   if Available (C) and not BookedRight(C)  then 
            		  		Available(C) := False; Score:= 1; -- Left stick allocated 
            			  	BookedLeft(C) := False; -- no more reason for booking left stick
							if Available(C + 1) and not BookedLeft(C + 1) then
            		  			Available(C + 1) := False; Score := 2;
					 			-- Right stick also allocated on the spot
						    else
								BookedRight(C + 1) := True;
						    end if;
					   else
							BookedLeft(C) := True; -- left stick reserved
					   end if;
					end if;
		 when 1 => if not BookedLeft(C + 1) and Available(C + 1)  then
            		   Available(C + 1) := False; Score := 2;
         		       -- Right stick now also allocated, no more reason for booking it 
					   BookedRight(C + 1) := False;
		 		   end if;
		 when 2 => null;
         end case;
		 if Score /= 2 then 
          		NotifyAll := False; requeue Wait; -- Java wait simulation
         end if;
	  end Get_LR;
     
      entry Wait(C : Id; Score : in out Cardinal) when NotifyAll is
      begin
		-- when NotifyAll is True, the eggshell model gives precedence to all queued tasks
		null; -- all queued tasks quit the monitor
      end Wait;

      procedure Release(C : Id) is
      begin
			Available(C) := True; Available(C + 1) := True;
			NotifyAll := (Wait'Count > 0); -- Java NotyfyAll simulation
      end Release;
	  
    end Sticks;
 	
   	procedure Get_LR(C : Id) is		
		Score :  Cardinal := 0;
		-- Score is the number of already allocated sticks"
	begin
		-- the following participate simulating the Java signalling semantics	
		while Score /= 2 loop 
			Sticks.Get_LR(C, Score);
		end loop;
	end Get_LR;
	
	procedure Release(C : Id) is
	begin Sticks.Release(C); end Release;
	
begin
	null;
	-- New_Line;
	-- Put_Line("*****   DINING PHILOSOPHERS WITH STICK ALLOCATION AS MANY AS AVAILABLE");
	-- Put_Line("*****   RELIABLE AND FAIR JAVA SOLUTION, BOOKING LEFT AND RIGHT STICK ");
	-- Put_Line("*****   AND GIVING PRECEDENCE TO ALREADY WAITING NEIGHBOURS ");
	-- Put_Line("*****   SIMULATED IN ADA WITH CONCURRENCY INSTRUMENTATION");
end Chop;
