-- dining philosophers paradigm. 

-- The chopsticks are allocated as many as available 
-- and the allocation is completed as soon as the missing chopsticks are released.
-- allocates two sticks if they are available
-- else allocate left stick if possible 
-- waits if both sticks have not been allocated
 
-- JAVA implementation with Notification objects
-- condition synchronisation provided by additional notification objects
-- Solution inspired by private semaphore schema

-- simulated with an Ada protected object 
--    which simulates Java concurrency semantics
-- the Java wait operation is here an entry call with a barrier
-- one Java synchronized method simulated by Get_LR and Wait

-- instrumented for concurrency denials ratio measurement

-- ********    reliable, deadlock free

with Global; use global;
package body Chop is

	type SticState is array(Id) of Boolean;
	subtype Cardinal is Integer range 0..2;
	type SticScore is array(Id) of Cardinal;
	
	protected Sticks is
		entry Get_LR(C : Id; Done : out Boolean); 
		procedure Release(C : in Id);
	private 
		entry Wait(Id)(C : in Id; Done : out Boolean);
		Available: SticState := (others => True);
		Score : SticScore := (others => 0); -- allocation state
		Requesting : SticState := (others => False);
		Notify : SticState := (others => False); -- Java signals simulation  
	end Sticks;
	
    protected body Sticks is
   
      entry  Get_LR(C : Id; Done : out Boolean) when True is
	  begin
		  	if Score(C) = 0 then -- this is the first time the call enters the monitor
				Requesting(C) := True;
				if Available (C) and  Available (C + 1) then 
					Available (C) := False; Available (C + 1) := False; 
					Score(C) := 2; -- Both sticks  allocated
					Requesting(C) := False; -- Successful first time 
				elsif Available (C) then
					Available (C) := False; Score(C):= 1; -- Left stick allocated	
				end if;
			end if; 		-- else the caller has been signalled and Requesting(C) = False
			Done := not Requesting(C); -- possibly False at first call, always true at second one
			if Requesting(C) then  
				Notify(C) := False; requeue Wait(C); -- Java wait simulation
			end if;		
	  end Get_LR;
     
      entry Wait(for C in Id)(C : Id; Done :  out Boolean) when Notify(C) is
      begin
			null;
      end Wait;

      procedure Release(C : Id) is
      begin
			Available(C) := True; Available(C + 1) := True; Score(C) := 0;
			-- now waiting philosophers are served preferentially and immediately
			-- if the left neighbour has already its left stick, it is serviced first
			if Requesting(C - 1) and Score(C - 1) = 1 then
				Available (C) := False; Score(C - 1) := 2; Requesting(C - 1) := False;
				Notify(C - 1) := True; -- the left neighbour is signalled
			end if;
			-- if the right neighbour requests its first stick, it is serviced immediately
			if Requesting(C + 1) then
				Available (C + 1) := False; Score(C + 1):= 1; -- Left stick allocated
				if Available (C + 2) then     -- right stick also allocated
					Available (C + 2) := False; Score(C + 1):= 2; 	Requesting(C + 1) := False; 
					Notify(C + 1) := True; -- the right neighbour is signalled			
				end if;
			end if;
      end Release;
	  
    end Sticks;
 	
   	procedure Get_LR(C : Id) is		
		AbleToEat :  Boolean := False; Score : Integer;
		-- AbleToEat is true when both sticks have been allocated
	begin
		-- the following simulates the Java queuing policy	
		while not AbleToEat loop
			Sticks.Get_LR(C, AbleToEat);
		end loop;
	end Get_LR;
	
	procedure Release(C : Id) is
	begin Sticks.Release(C); end Release;
	
begin
	null; -- New_Line;
	-- Put_Line("*****   DINING PHILOSOPHERS WITH STICK ALLOCATION AS MANY AS AVAILABLE");
	-- Put_Line("*****   RELIABLE  JAVA NOTIFICATION OBJECTS SOLUTION : SIMULATED IN ADA");
	-- Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
