-- dining philosophers paradigm. 

-- The chopsticks are allocated as many as available 
-- and the allocation is completed as soon as the missing chopsticks are released.
-- allocates two sticks if they are available
-- else allocate left stick if possible 
-- waits if both sticks have not been allocated
 
-- JAVA implementation with Lock and conditions of J2SE/JDK 1.5
-- condition synchronisation provided by additional condition objects
-- Solution when a monitor allows several named condition variable

-- simulated with an Ada protected object 
--    which simulates Java concurrency semantics
-- the Java condition.await operation is here an entry call with a barrier
-- one Java synchronized method simulated by Get_LR  and condition entries

-- instrumented for concurrency denials ratio measurement

-- ********    reliable, deadlock free

with Global; use global;

package body Chop is

	type SticState is array(Id) of Boolean;
	subtype Cardinal is Integer range 0..2;
	type SticScore is array(Id) of Cardinal;
	
	protected Sticks is
		entry Get_LR(C : Id; Done : out Boolean); 
		procedure Release(C : in Id);
	private 
		entry LeftAllocated(Id)(C : in Id; Done : out Boolean);
		entry RightAllocated(Id)(C : in Id; Done : out Boolean);
		Available: SticState := (others => True); 
		BookedRight: SticState := (others => False);
		Score : SticScore := (others => 0); -- allocation state of a philosopher
		SignalLeft : SticState := (others => False); -- Java signals simulation  
		SignalRight : SticState := (others => False); -- Java signals simulation
	end Sticks;
	
    protected body Sticks is
   
      entry  Get_LR(C : Id; Done : out Boolean) when True is
	  begin
		  	if Score(C) = 0 then -- this is the first time the call enters the monitor
				if not Available(C) or BookedRight(C) then 
					Done := False; SignalLeft(C) := False; requeue LeftAllocated(C); 
					-- this above has simulated Java LeftAllocated(C).await
				else
					Available (C) := False; BookedRight(C + 1) := True; 
					Score(C) := 1; -- Left sticks allocated
				end if;
			end if; 
		  	if Score(C) = 1 then -- C has its left stick allocated
				if not Available(C + 1) then
					Done := False; SignalRight(C) := False; requeue RightAllocated(C); 
					-- this above has simulated Java RightAllocated(C).await
				else
					BookedRight(C + 1) := False;  -- no more reason for booking right stick
					Available (C + 1) := False; Score(C):= 2; -- Right stick allocated	
					Done := True; -- both sticks are allocated
				end if; -- back to the Java ready thread queue
			end if; 
	  end Get_LR;
     
      entry LeftAllocated(for C in Id)(C : Id; Done :  out Boolean) when SignalLeft(C) is
      begin
			null;
      end LeftAllocated; -- back to the Java ready thread queue 

      entry RightAllocated(for C in Id)(C : Id; Done :  out Boolean) when SignalRight(C) is
      begin
			null;
      end RightAllocated; -- back to the Java ready thread queue

     procedure Release(C : Id) is
      begin
			Available(C) := True; Available(C + 1) := True; Score(C) := 0;
			-- now waiting philosophers are served preferentially and immediately
			-- if the left neighbour has already its left stick, it is serviced first
			if BookedRight(C) then 
				-- the left neighbour requests its right stick (also Score(C - 1) = 1)
				SignalRight(C - 1) := True; -- the left neighbour C - 1 is signalled
			end if;
			-- if the right neighbour requests its first stick, it is serviced immediately
			SignalLeft(C + 1) := True; -- if the right neighbour is waiting, it is signalled
	  end Release; -- back to the Java ready thread queue 
	  
    end Sticks;
 	
   	procedure Get_LR(C : Id) is	
		AbleToEat :  Boolean := False; 
		-- AbleToEat is true when both sticks have been allocated
	begin
		-- the following simulates the Java queuing policy	
		while not AbleToEat loop
			Sticks.Get_LR(C, AbleToEat);
		end loop;
	end Get_LR;
	
	procedure Release(C : Id) is
	begin Sticks.Release(C); end Release;
	
begin
	null; -- New_Line;
	-- Put_Line("*****   DINING PHILOSOPHERS WITH STICK ALLOCATION AS MANY AS AVAILABLE");
	-- Put_Line("*****   RELIABLE JAVA JDK 1.5  CONDITION OBJECTS SOLUTION : SIMULATED IN ADA");
	-- Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
