-- dining philosophers paradigm.

-- The chopsticks are allocated as many as available 
-- and the allocation is completed as soon as the missing chopsticks are released.
 
-- JAVA implementation simulated with an Ada protected object 
-- simulates Java concurrency semantics
-- the Java wait operation is here an entry call with a barrier
-- the Java synchronized method simulated by Get_LR 
-- allocates two sticks if they are available
-- else allocate left stick if possible 
-- waits if both sticks have not been allocated

-- If the right stick is not available it is booked
-- a booked left stick is not allocated, this avoids deadlock

-- instrumented for concurrency denials ratio measurement

-- ********    reliable, deadlock free

with Global; use global;
with Monitoring; use Monitoring;
with Ada.Text_IO;  use Ada.Text_IO;
package body Chop is

	type SticState is array(Id) of Boolean;
	type Cardinal is new Integer range 0..2;
	
	protected Sticks is
		entry Get_LR(C : Id; Score : in out Cardinal); 
		procedure Release(C : in Id);
		-- Score is the number of already allocated sticks
	private 
		entry Wait(C : Id; Score : in out Cardinal);
		Available: SticState := (others => True);
		Booked : SticState := (others => False);
		NotifyAll : Boolean := False; -- Java signal simulation  
	end Sticks;
	
    protected body Sticks is
   
      entry  Get_LR(C : Id; Score : in out Cardinal) when True is
		begin
			case Score is
        	    when 0 => if not Booked(C) and Available (C) then 
            		  		Available(C) := False;  
					  		Score:= 1; -- Left stick allocated 
            			  	if Available(C + 1) then
            		  			Available(C + 1) := False; Score := 2;
					 			-- Right stick also allocated on the spot
						    else
								Booked(C + 1) := True;
						    end if;
						  end if;
		 	 	when 1 => if Available(C + 1) then
            		  		Available(C + 1) := False; Score := 2;
         		      		-- Right stick now also allocated; 
							Booked(C + 1) := False;
		 		    	  end if;
		 		when 2 => null;
        	end case;
			if Score /= 2 then 
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
          		NotifyAll := False; requeue Wait; -- Java wait simulation
			else
				Monitoring.InstrumentGetSuccessful(C); -- counts the success once
        	end if;
		end Get_LR;
     
      entry Wait(C : Id; Score : in out Cardinal) when NotifyAll is
      begin
		-- when NotifyAll is True, the eggshell model gives precedence to all queued tasks
		null; -- all queued tasks quit the monitor
      end Wait;

      procedure Release(C : Id) is
      begin
			Available(C) := True; Available(C + 1) := True;
			NotifyAll := (Wait'Count > 0); -- Java NotyfyAll simulation
			Monitoring.InstrumentRelease(C);
      end Release;
	  
    end Sticks;
 	
   	procedure Get_LR(C : Id) is		
		Score :  Cardinal := 0;
		-- Score is the number of already allocated sticks"
	begin
		ChopStart(C); -- starting service time for this allocation
		-- the following participate simulating the Java signalling semantics	
		while Score /= 2 loop 
			Sticks.Get_LR(C, Score);
			Requesting_LR(C, Integer(Score)); 
		end loop;
		ChopEnd(C); -- ending  service time for this allocation
	end Get_LR;
	
	procedure Release(C : Id) is
	begin Sticks.Release(C); end Release;
	
begin
	New_Line;
	Put_Line("*****   DINING PHILOSOPHERS WITH STICK ALLOCATION AS MANY AS AVAILABLE");
	Put_Line("*****   RELIABLE JAVA SOLUTION, BOOKING RIGHT STICK : SIMULATED IN ADA");
	Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
