-- dining philosophers paradigm.

-- The chopsticks are allocated as many as available 
-- and the allocation is completed as soon as the missing chopsticks are released.
 
-- ADA solution using the eggshell semantic of Ada protected object 

-- The use of Ada entry families and requeue statement provides 
-- an elegant and correct (i.e., no deadlock, no starvation) solution 
-- with at most one task waiting at an entry, with requeuing just once 
-- (no pseudo busy waiting while scanning entry queues)
--  and with a simple Boolean variable as entry barrier. 

-- instrumented for concurrency denials ratio measurement

-- ***************     reliable, no deadlock, deterministic

with Global; use global;
with Monitoring; use Monitoring;
with Ada.Text_IO;  use Ada.Text_IO;
package body Chop is

	type SticState is array(Id) of Boolean;
	type Flush_Array is array(Id) of Integer; 

	protected Sticks is
		entry Get_LR(C : in Id);  
		procedure Release(C : in Id);
	private 
		entry Assign_LR(Id); -- entry family
		entry Get_R(Id);  -- entry family
		entry Assign_R(Id); -- entry family
		Available : SticState := (others => True); -- Stick availability
		Flush_Count_LR, Flush_Count_R : Flush_Array := (others => 0);
	end Sticks;
	
	protected body Sticks is   

		entry Get_LR(C : in Id) when True is
		begin
			if Available (C) then
				Available (C) := False;
				requeue Get_R(C);
			else
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
				requeue Assign_LR(C);
			end if;
		end Get_LR;

 		entry Assign_LR(for C in Id) when Flush_Count_LR(C) > 0 is
		begin
			Flush_Count_LR(C) := Flush_Count_LR(C) - 1;
			if Available (C) then 		-- Left stick allocated 
				Available (C) := False;
				requeue Get_R(C);
			else 
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal
				requeue Assign_LR(C) ;
			end if; 
		end Assign_LR;    

		entry Get_R(for C  in Id) when True is
		begin
			if Available (C + 1) then -- stick allocated as a right stick
				Available (C + 1) := False;
				Monitoring.InstrumentGetSuccessful(C); -- counts the success once
			else
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
				requeue Assign_R(C);
			end if;	
		end Get_R;

 		entry Assign_R(for C in Id) when Flush_Count_R(C) > 0 is
		begin
			Flush_Count_R(C) := Flush_Count_R(C) - 1;
			if Available (C + 1) then -- stick allocated as a right stick
				Available (C + 1) := False;
				Monitoring.InstrumentGetSuccessful(C); -- counts the success once
			else
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
				requeue Assign_R(C);
			end if; 
		end Assign_R; 
   
		procedure Release(C : Id) is
		begin
			Available(C) := True; Available(C + 1) := True; 
			Monitoring.InstrumentRelease(C);
			Flush_Count_R(C- 1) := Assign_R(C - 1)'Count; -- C- 1 waiting for a right stick, C
			Flush_Count_LR(C + 1) := Assign_LR(C + 1)'Count; -- C+ 1 waiting for a left stick, C + 1
		end Release;
	end Sticks;

	procedure Get_LR(C : Id) is	
	begin
		ChopStart(C); -- starting service time for sticks allocation 
		Sticks.Get_LR(C); 
		ChopEnd(C); -- ending  service time for sticks allocation
	end Get_LR;
	procedure Release(C : Id) is begin Sticks.Release(C); end Release;

begin
	New_Line;
	Put_Line("*****   DINING PHILOSOPHERS WITH STICK ALLOCATION AS MANY AS AVAILABLE");
	Put_Line("*****   RELIABLE AND DETERMINISTIC ADA SOLUTION USING REQUEUE AND FAMILIES");
	Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
